import pytest
import mock
from mock import PropertyMock, call

from tap_mambu.helpers.generator_processor_pairs import get_generator_processor_for_stream, \
    get_generator_processor_pairs, get_stream_subtypes
from tap_mambu.sync import sync_endpoint, sync_all_streams
from tap_mambu.tap_generators.child_generator import ChildGenerator
from .helpers import IsInstanceMockMatcher
from .constants import config_json


@mock.patch("tap_mambu.helpers.generator_processor_pairs.LoanAccountsLMGenerator")
@mock.patch("tap_mambu.helpers.generator_processor_pairs.LoanAccountsADGenerator")
@mock.patch("tap_mambu.helpers.generator_processor_pairs.LoanAccountsProcessor")
def test_sync_endpoint(mock_loan_accounts_processor, mock_loan_accounts_ad_generator,
                       mock_loan_accounts_lm_generator):
    type(mock_loan_accounts_lm_generator.return_value).type = PropertyMock(return_value="LoanAccountsLMGenerator")
    type(mock_loan_accounts_ad_generator.return_value).type = PropertyMock(return_value="LoanAccountsADGenerator")
    client = "client"
    catalog = "catalog"
    state = "state"
    stream_name = "loan_accounts"
    sub_type = "sub_type"
    config = "config"
    sync_endpoint(client=client, catalog=catalog, state=state, stream_name=stream_name,
                  sub_type=sub_type, config=config)
    mock_loan_accounts_lm_generator.assert_called_once_with(client=client, state=state, stream_name=stream_name,
                                                            sub_type=sub_type, config=config)
    mock_loan_accounts_ad_generator.assert_called_once_with(client=client, state=state, stream_name=stream_name,
                                                            sub_type=sub_type, config=config)
    mock_loan_accounts_processor.assert_called_once_with(client=client, state=state, stream_name=stream_name,
                                                         sub_type=sub_type, config=config, catalog=catalog,
                                                         generators=[IsInstanceMockMatcher("LoanAccountsLMGenerator"),
                                                                     IsInstanceMockMatcher("LoanAccountsADGenerator")])


@mock.patch("tap_mambu.sync.sync_endpoint")
@mock.patch("tap_mambu.sync.update_currently_syncing")
@mock.patch("tap_mambu.sync.singer.get_currently_syncing")
@mock.patch("tap_mambu.sync.get_generator_processor_for_stream")
@mock.patch("tap_mambu.sync.get_selected_streams")
@mock.patch("tap_mambu.sync.get_timezone_info")
def test_sync_all_streams_flow(mock_get_timezone_info,
                               mock_get_selected_streams,
                               mock_get_generator_processor_for_stream,
                               mock_singer_get_currently_syncing,
                               mock_update_currently_syncing,
                               mock_sync_endpoint):
    available_stream_pairs = get_generator_processor_pairs()
    selected_streams = list(available_stream_pairs.keys())
    child_streams = [stream_name for stream_name, pairs in available_stream_pairs.items()
                     for generator in pairs[0]
                     if issubclass(generator, ChildGenerator)]

    client = "client",
    catalog = "catalog",
    state = {'currently_syncing': 'activities'}

    mock_get_selected_streams.return_value = selected_streams
    mock_get_generator_processor_for_stream.side_effect = lambda x: get_generator_processor_for_stream(x)
    mock_singer_get_currently_syncing.return_value = 'activities'

    sync_all_streams(client, config_json, catalog, state)

    mock_get_selected_streams.assert_called_with(catalog)
    mock_singer_get_currently_syncing.assert_called_with(state)
    mock_get_generator_processor_for_stream.assert_has_calls([call(stream_name) for stream_name in selected_streams])
    assert mock_get_generator_processor_for_stream.call_count == len(selected_streams)

    selected_streams_count = 0
    for stream_name in selected_streams:
        # did in this way because gl_accounts counts as 5 streams due to its subtypes
        selected_streams_count += len(get_stream_subtypes(stream_name))
        if stream_name not in child_streams:
            state['currently_syncing'] = stream_name
            mock_update_currently_syncing.assert_any_call(state, stream_name)
            mock_update_currently_syncing.assert_any_call(state, None)

            for sub_type in get_stream_subtypes(stream_name):
                mock_sync_endpoint.assert_has_calls([call(client=client,
                                                          catalog=catalog,
                                                          state=state,
                                                          stream_name=stream_name,
                                                          sub_type=sub_type,
                                                          config=config_json)])

    assert mock_update_currently_syncing.call_count == (selected_streams_count - len(child_streams)) * 2
    assert mock_sync_endpoint.call_count == selected_streams_count - len(child_streams)


@mock.patch("tap_mambu.sync.singer.get_currently_syncing")
@mock.patch("tap_mambu.sync.get_selected_streams")
@mock.patch("tap_mambu.sync.get_timezone_info")
def test_sync_all_streams_no_selected_streams(mock_get_timezone_info,
                                              mock_get_selected_streams,
                                              mock_singer_get_currently_syncing):
    mock_get_selected_streams.return_value = set()
    sync_all_streams('client', {}, 'catalog', 'state')

    mock_get_selected_streams.assert_called_with('catalog')
    mock_singer_get_currently_syncing.assert_not_called()


@mock.patch("tap_mambu.sync.should_sync_stream")
@mock.patch("tap_mambu.sync.get_selected_streams")
@mock.patch("tap_mambu.sync.get_timezone_info")
def test_sync_all_streams_flow_child_stream(mock_get_timezone_info,
                                            mock_get_selected_streams,
                                            mock_should_sync_stream):
    streams_without_child = []
    child_streams = []

    for stream_name, pairs in get_generator_processor_pairs().items():
        for generator in pairs[0]:
            if issubclass(generator, ChildGenerator):
                child_streams.append(stream_name)
            else:
                streams_without_child.append(stream_name)

    all_streams = streams_without_child + child_streams
    mock_get_selected_streams.return_value = all_streams
    mock_should_sync_stream.return_value = (None, None)

    sync_all_streams('client', {}, 'catalog', {})

    for stream_name in streams_without_child:
        mock_should_sync_stream.assert_any_call(all_streams,
                                                None,
                                                stream_name)

    for stream_name in child_streams:
        with pytest.raises(AssertionError):
            mock_should_sync_stream.assert_called_with(all_streams,
                                                       None,
                                                       stream_name)



from . import setup_generator_base_test


def test_clients_generator():
    generators = setup_generator_base_test("clients")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == "clients:search"
    assert generator.endpoint_bookmark_field == "lastModifiedDate"
    assert generator.endpoint_sorting_criteria == {
            "field": "lastModifiedDate",
            "order": "ASC"
        }
    assert generator.endpoint_filter_criteria == [
            {
                "field": "lastModifiedDate",
                "operator": "AFTER",
                "value": '2021-05-31T17:00:00.000000Z'
            }
        ]



import json
import pytest
from mock import Mock, patch, call

from mambu_tests.helpers import ClientMock, MultithreadedOffsetGeneratorFake
from tap_mambu.helpers.hashable_dict import HashableDict
from tap_mambu.tap_generators.multithreaded_offset_generator import MultithreadedOffsetGenerator


def test_init_variable_params():
    mock_client = ClientMock()
    mock_config = Mock()
    mock_config.start_date = "2019-11-25T00:00:00.000000Z"

    generator = MultithreadedOffsetGeneratorFake(config=mock_config)

    assert generator.artificial_limit == mock_client.page_size
    assert generator.limit == mock_client.page_size + generator.overlap_window
    assert generator.batch_limit == generator.max_threads * generator.client.page_size + generator.overlap_window


def test_check_and_get_set_reunion():
    limit = 100
    overlap_window = 10
    total_records_no = limit + overlap_window

    generator = MultithreadedOffsetGeneratorFake()
    generator.overlap_window = limit
    generator.artificial_limit = overlap_window

    # test when a and b have the same values on the overlap window indexes
    a = [json.dumps({'encoded_key': f'0-{record_no}-test'}) for record_no in range(0, total_records_no)]
    b = a[-overlap_window:] + [json.dumps({'encoded_key': f'1-{record_no}-test'}) for record_no in range(0, limit)]
    assert generator.check_and_get_set_reunion(set(a), set(b), overlap_window) == set(a).union(b)

    # test when a and b don't overlap 100% but the error it's manageable
    a = [json.dumps({'encoded_key': f'0-{record_no}-test'}) for record_no in range(0, total_records_no)]
    b = a[-(overlap_window - 5):] + [json.dumps({'encoded_key': f'1-{record_no}-test'})
                                     for record_no in range(0, limit)]
    assert generator.check_and_get_set_reunion(set(a), set(b), overlap_window) == set(a).union(b)

    # test when a and b don't overlap at all
    with pytest.raises(RuntimeError):
        a = {json.dumps({'encoded_key': f'0-{record_no}-test'}) for record_no in range(0, total_records_no)}
        b = {json.dumps({'encoded_key': f'1-{record_no}-test'}) for record_no in range(0, total_records_no)}
        generator.check_and_get_set_reunion(a, b, overlap_window)


@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator.stop_all_request_threads")
@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator.check_and_get_set_reunion",
       side_effect=RuntimeError)
def test_error_check_and_fix(mock_check_and_get_set_reunion, mock_stop_all_request_threads):
    generator = MultithreadedOffsetGeneratorFake()

    mock_final_buffer = {f'0-{record_no}'
                         for record_no in range(0, generator.client.page_size + generator.overlap_window)}
    mock_temp_buffer = {f'1-{record_no}'
                        for record_no in range(0, generator.client.page_size)}

    # test if the method raises RuntimeError if it can't correct the batches
    with pytest.raises(RuntimeError):
        generator.error_check_and_fix(mock_final_buffer, mock_temp_buffer, [])

    mock_check_and_get_set_reunion.assert_called_with(mock_final_buffer, mock_temp_buffer, generator.client.page_size)
    mock_stop_all_request_threads.assert_called_with([])


@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator.fetch_batch_continuously")
def test_iter_flow(mock_fetch_batch_continuously):
    generator = MultithreadedOffsetGeneratorFake()

    mock_fetch_batch_continuously.assert_not_called()
    assert generator.fetch_batch_thread is None

    iter(generator)
    mock_fetch_batch_continuously.assert_called_once()
    assert generator.fetch_batch_thread is not None


def test_next_flow():
    test_records = [{"encoded_key": "1", "creation_date": '2020-01-01T00:00:000000Z'},
                    {"encoded_key": "2", "creation_date": '2020-01-01T00:00:000000Z'},
                    {"encoded_key": "3", "creation_date": '2020-01-01T00:00:000000Z'}, ]
    generator = MultithreadedOffsetGeneratorFake()
    generator.buffer = list(test_records)

    for record in test_records:
        assert next(generator) == record

    with pytest.raises(StopIteration):
        generator.end_of_file = True
        next(generator)


@patch("tap_mambu.tap_generators.multithreaded_offset_generator.time.sleep")
def test_stop_all_request_threads(mock_time_sleep):
    def create_future(side_effect=None):
        future = Mock()
        future.cancel = Mock()
        future.done = Mock()
        if side_effect:
            future.done.side_effect = side_effect
        else:
            future.done.return_value = True
        return future

    finished_futures = [create_future() for _ in range(0, 100)]
    not_finished_futures = [create_future(side_effect=[False, True]) for _ in range(0, 5)]

    MultithreadedOffsetGeneratorFake.stop_all_request_threads(finished_futures + not_finished_futures)

    assert mock_time_sleep.call_count == len(not_finished_futures)

    # test if the cancel and done methods are called the correct amount of times
    for finished_future in finished_futures:
        assert finished_future.cancel.call_count == 1
        assert finished_future.done.call_count == 1

    # test if the cancel and done methods are called the correct amount of times
    for not_finished_future in not_finished_futures:
        assert not_finished_future.cancel.call_count == 1
        assert not_finished_future.done.call_count == 2


@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator._all_fetch_batch_steps")
@patch("tap_mambu.tap_generators.multithreaded_offset_generator.time.sleep")
def test_fetch_batch_continuously_first_call(mock_time_sleep, mock_all_fetch_batch_steps):
    mock_all_fetch_batch_steps.return_value = False

    generator = MultithreadedOffsetGeneratorFake()
    assert generator.end_of_file is False
    assert len(generator.buffer) == 0

    # test if the first run of the method does call only the _all_fetch_batch_steps method
    generator.fetch_batch_continuously()
    mock_all_fetch_batch_steps.assert_called_once()
    mock_time_sleep.assert_called_once()
    assert generator.end_of_file is True


@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator._all_fetch_batch_steps")
@patch("tap_mambu.tap_generators.multithreaded_offset_generator.time.sleep")
def test_fetch_batch_continuously_multiple_calls(mock_time_sleep, mock_all_fetch_batch_steps):
    mock_all_fetch_batch_steps.side_effect = [True, True, False]

    generator = MultithreadedOffsetGeneratorFake()
    generator.fetch_batch_continuously()

    # test the flow if multiple while iterations are needed
    mock_all_fetch_batch_steps.assert_called()
    assert mock_all_fetch_batch_steps.call_count == 3
    assert mock_time_sleep.call_count == 3
    assert generator.end_of_file is True


def test_preprocess_record_one_record():
    mock_record = {'encoded_key': 'test', 'test_field': 'value'}

    generator = MultithreadedOffsetGeneratorFake()
    assert generator.buffer == []

    record = generator.preprocess_record(mock_record)
    assert record == mock_record
    assert generator.buffer == [mock_record]


def test_preprocess_record_multiple_records():
    mock_records = [{'encoded_key': 'test', 'test_field': f'value_{no}'} for no in range(10)]

    generator = MultithreadedOffsetGeneratorFake()
    assert generator.buffer == []

    for idx, mock_record in enumerate(mock_records):
        record = generator.preprocess_record(mock_record)
        assert record == mock_records[idx]
    assert generator.buffer == mock_records


@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator.preprocess_record")
def test_preprocess_batches_flow(mock_preprocess_record):
    mock_records = [{'encoded_key': 'test', 'test_field': f'value_{no}'} for no in range(100)]
    mock_last_batch_set = set(f'{{"encoded_key":"test_value{no}"}}' for no in range(50))

    generator = MultithreadedOffsetGeneratorFake()
    assert generator.last_batch_set == set()

    generator.last_batch_set = mock_last_batch_set
    generator.preprocess_batches(mock_records)

    mock_preprocess_record.assert_has_calls([call(record) for record in mock_records])
    assert mock_preprocess_record.call_count == len(mock_records)
    assert generator.last_batch_size == len(mock_last_batch_set)


@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator.preprocess_batches")
@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator.collect_batches")
@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator.queue_batches")
def test_all_fetch_batch_steps_flow(mock_queue_batches, mock_collect_batches, mock_preprocess_batches):
    generator = MultithreadedOffsetGeneratorFake()

    # test simple flow
    mock_collect_batches.return_value = (set('test_value'), False)
    fetch_output = generator._all_fetch_batch_steps()
    assert fetch_output is True

    # test when stop_iteration is True
    mock_collect_batches.return_value = (set('test_value'), True)
    fetch_output = generator._all_fetch_batch_steps()
    assert fetch_output is False

    # test when final_buffer is empty
    mock_collect_batches.return_value = (set(), False)
    fetch_output = generator._all_fetch_batch_steps()
    assert fetch_output is False

    # test when stop_iteration is True and final_buffer is empty
    mock_collect_batches.return_value = (set(), True)
    fetch_output = generator._all_fetch_batch_steps()
    assert fetch_output is False

    assert mock_queue_batches.call_count == 4
    assert mock_collect_batches.call_count == 4
    assert mock_preprocess_batches.call_count == 4


@patch.object(MultithreadedOffsetGenerator, 'prepare_batch',
              side_effect=MultithreadedOffsetGenerator.prepare_batch,
              autospec=True)
@patch("tap_mambu.tap_generators.multithreaded_offset_generator.MultithreadedRequestsPool.queue_request")
def test_queue_batches(mock_queue_request,
                       mock_prepare_batch):
    mock_client = ClientMock()
    mock_endpoint_path = 'test_endpoint_path'
    mock_endpoint_api_method = 'POST'
    mock_endpoint_api_version = 'v2'
    mock_endpoint_api_key_type = 'test_endpoint_api_key_type'
    mock_endpoint_body = {'filterCriteria': {"field": "creationDate",
                                             "operator": "BETWEEN",
                                             "value": '2020-01-01T12:00:00.000000Z',
                                             "secondValue": '2020-02-01T03:01:20.000000Z'},
                          'sortingCriteria': {"field": "encodedKey",
                                              "order": "ASC"}
                          },
    mock_params = {'paginationDetails': 'OFF',
                   'detailsLevel': 'FULL'}
    mock_overlap_window = 20
    mock_batch_limit = 4000
    mock_artificial_limit = mock_client.page_size

    mock_queue_request.side_effect = [Mock() for _ in range(0, mock_batch_limit, mock_artificial_limit)]

    generator = MultithreadedOffsetGeneratorFake(client=mock_client)
    generator.overlap_window = mock_overlap_window
    generator.batch_limit = mock_batch_limit
    generator.endpoint_path = mock_endpoint_path
    generator.endpoint_api_method = mock_endpoint_api_method
    generator.endpoint_api_version = mock_endpoint_api_version
    generator.endpoint_api_key_type = mock_endpoint_api_key_type
    generator.endpoint_body = mock_endpoint_body
    generator.params = mock_params
    generator.limit = mock_client.page_size + mock_overlap_window

    # +1 because the while loop condition is <=
    while_upper_limit = mock_batch_limit // (mock_client.page_size + mock_overlap_window) + 1

    assert generator.offset == 0
    features = generator.queue_batches()

    assert len(features) == while_upper_limit

    mock_params['offset'] = 0
    mock_params['limit'] = mock_artificial_limit + mock_overlap_window
    calls = []
    for offset in range(0, while_upper_limit):
        calls.append(call(mock_client, 'test_stream', mock_endpoint_path, mock_endpoint_api_method,
                          mock_endpoint_api_version, mock_endpoint_api_key_type, mock_endpoint_body, dict(mock_params)))
        mock_params['offset'] += mock_artificial_limit

    # test if the queue_request method is called using the correct offset
    mock_queue_request.assert_has_calls(calls)

    # test if the methods are called the correct amount of times
    assert mock_prepare_batch.call_count == while_upper_limit
    assert mock_queue_request.call_count == while_upper_limit
    # used only artificial_limit because the offset is increased only by the artificial_limit value (without overlap)
    assert generator.offset == while_upper_limit * mock_artificial_limit


@patch("tap_mambu.tap_generators.multithreaded_offset_generator.transform_json")
@patch("tap_mambu.tap_generators.multithreaded_offset_generator.time.sleep")
def test_collect_batches_thread_not_done(mock_time_sleep, mock_transform_json):
    def create_mock_future_pending():
        mock_future_pending = Mock()
        mock_future_pending.done = Mock()
        mock_future_pending.done.side_effect = [False, False, True]
        return mock_future_pending

    mock_future_done = Mock()
    mock_future_done.done = Mock()
    mock_future_done.done.return_value = True

    generator = MultithreadedOffsetGeneratorFake()

    mock_futures = [mock_future_done for _ in
                    range(0, generator.batch_limit - generator.artificial_limit,
                          generator.artificial_limit)] + \
                   [create_mock_future_pending()]

    generator.collect_batches(mock_futures)

    # test if the sleep function is called when the request isn't finished
    assert mock_time_sleep.call_count == 2

    # just a check to make sure the algorithm passed over the while loop
    mock_transform_json.assert_called()


@patch.object(MultithreadedOffsetGenerator, 'error_check_and_fix',
              side_effect=MultithreadedOffsetGenerator.error_check_and_fix,
              autospec=True)
@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator.stop_all_request_threads")
@patch("tap_mambu.tap_generators.multithreaded_offset_generator.MultithreadedOffsetGenerator.transform_batch")
@patch("tap_mambu.tap_generators.multithreaded_offset_generator.transform_json")
def test_collect_batches(mock_transform_json,
                         mock_transform_batch,
                         mock_stop_all_request_threads,
                         mock_error_check_and_fix):
    mock_batch_limit = 4000
    mock_artificial_limit = 200
    mock_overlap_window = 20
    mock_client = ClientMock(page_size=200)

    generator = MultithreadedOffsetGeneratorFake(client=mock_client)
    generator.artificial_limit = mock_artificial_limit
    generator.batch_limit = mock_batch_limit
    generator.overlap_window = mock_overlap_window

    # generate fake data as if they were extracted from the API
    mock_records = [[HashableDict({'encoded_key': f'0-{record_no}'})
                     for record_no in range(0, mock_client.page_size + mock_overlap_window)], ]
    for batch_no in range(1, (mock_batch_limit // mock_artificial_limit) - 1):
        mock_batch = mock_records[batch_no - 1][-mock_overlap_window:]
        for record_no in range(0, mock_client.page_size):
            mock_batch.append(HashableDict({'encoded_key': f'{batch_no}-{record_no}'}))
        mock_records.append(mock_batch)
    mock_records.append([])

    mock_transform_batch.side_effect = mock_records

    mock_features = [Mock() for _ in range(0, mock_batch_limit, mock_artificial_limit)]
    buffer, stop_iteration = generator.collect_batches(mock_features)

    assert stop_iteration is True
    assert all(record in buffer for batch in mock_records[:-1] for record in batch)

    assert mock_transform_json.call_count == mock_batch_limit / mock_artificial_limit
    assert mock_transform_batch.call_count == mock_batch_limit / mock_artificial_limit
    # -2 because the first call is skipped and the last future value is an empty list
    assert mock_error_check_and_fix.call_count == (mock_batch_limit / mock_artificial_limit) - 2
    mock_stop_all_request_threads.assert_called_once()



import abc

from .generator import TapGenerator
from ..helpers import get_bookmark
from ..helpers.datetime_utils import str_to_localized_datetime, datetime_to_local_str


class LoanAccountsGenerator(TapGenerator):
    @abc.abstractmethod
    def _init_endpoint_config(self):
        super(LoanAccountsGenerator, self)._init_endpoint_config()
        self.endpoint_path = "loans:search"
        self.endpoint_sorting_criteria = {
            "field": "id",
            "order": "ASC"
        }


class LoanAccountsLMGenerator(LoanAccountsGenerator):
    def _init_endpoint_config(self):
        super()._init_endpoint_config()
        self.endpoint_bookmark_field = "lastModifiedDate"
        self.endpoint_filter_criteria = [
            {
                "field": "lastModifiedDate",
                "operator": "AFTER",
                "value": datetime_to_local_str(str_to_localized_datetime(
                    get_bookmark(self.state, self.stream_name, self.sub_type, self.start_date)))
            }
        ]


class LoanAccountsADGenerator(LoanAccountsGenerator):
    def _init_endpoint_config(self):
        super()._init_endpoint_config()
        self.endpoint_bookmark_field = "lastAccountAppraisalDate"
        self.endpoint_filter_criteria = [
            {
                "field": "lastAccountAppraisalDate",
                "operator": "AFTER",
                "value": datetime_to_local_str(str_to_localized_datetime(
                    get_bookmark(self.state, self.stream_name, self.sub_type, self.start_date)))
            }
        ]



from . import setup_generator_base_test


def test_deposit_cards_generator_endpoint_config_init():
    generators = setup_generator_base_test("cards")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_parent_id == "TEST"
    assert generator.endpoint_path == 'deposits/TEST/cards'
    assert generator.endpoint_api_method == "GET"



import json
import threading
import time
from datetime import datetime

from mock import Mock, patch, call
from pytz import timezone

from mambu_tests.helpers import ClientMock, MultithreadedBookmarkGeneratorFake
from tap_mambu.helpers.hashable_dict import HashableDict
from tap_mambu.tap_generators.multithreaded_bookmark_generator import MultithreadedBookmarkGenerator


def test_set_intermediary_bookmark():
    mock_record_bookmark_value = '2020-01-01T00:00:00.000000Z'

    generator = MultithreadedBookmarkGeneratorFake()

    # test if the initial values aren't modified
    assert generator.endpoint_intermediary_bookmark_value is None
    assert generator.endpoint_intermediary_bookmark_offset == 0

    # check if the endpoint intermediary bookmark value and offset are set for a "correct" bookmark value
    generator.set_intermediary_bookmark(mock_record_bookmark_value)
    assert generator.endpoint_intermediary_bookmark_value == mock_record_bookmark_value
    assert generator.endpoint_intermediary_bookmark_offset == 1

    # check if the endpoint intermediary bookmark value isn't set with a value lower than the already saved one
    generator.set_intermediary_bookmark('2005-01-01T00:00:00.000000Z')
    assert generator.endpoint_intermediary_bookmark_value == mock_record_bookmark_value
    assert generator.endpoint_intermediary_bookmark_offset == 1

    # check if the endpoint intermediary bookmark value and offset correctly set when multiple records have
    # the same bookmark field value
    for _ in range(3):
        generator.set_intermediary_bookmark(mock_record_bookmark_value)
    assert generator.endpoint_intermediary_bookmark_value == mock_record_bookmark_value
    assert generator.endpoint_intermediary_bookmark_offset == 4

    # check if the endpoint intermediary bookmark value is properly set and the offset is reset
    mock_record_bookmark_value = '2021-11-11T00:00:00.000000Z'
    generator.set_intermediary_bookmark(mock_record_bookmark_value)
    assert generator.endpoint_intermediary_bookmark_value == mock_record_bookmark_value
    assert generator.endpoint_intermediary_bookmark_offset == 1


@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator."
       "MultithreadedBookmarkGenerator._all_fetch_batch_steps")
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.time.sleep")
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.MultithreadedBookmarkGenerator.prepare_batch_params")
def test_fetch_batch_continuously_first_call(mock_prepare_batch_params, mock_time_sleep, mock_all_fetch_batch_steps):
    mock_all_fetch_batch_steps.return_value = False

    generator = MultithreadedBookmarkGeneratorFake()
    assert generator.end_of_file is False
    assert len(generator.buffer) == 0

    # test if the first run of the method does call only the _all_fetch_batch_steps method
    generator.fetch_batch_continuously()
    mock_prepare_batch_params.assert_not_called()
    mock_all_fetch_batch_steps.assert_called_once()
    mock_time_sleep.assert_not_called()
    assert generator.end_of_file is True


@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator."
       "MultithreadedBookmarkGenerator._all_fetch_batch_steps")
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.MultithreadedBookmarkGenerator.prepare_batch_params")
def test_fetch_batch_continuously_multiple_calls(mock_prepare_batch_params, mock_all_fetch_batch_steps):
    mock_all_fetch_batch_steps.side_effect = [True, True, False]

    generator = MultithreadedBookmarkGeneratorFake()
    generator.fetch_batch_continuously()

    # test the flow if multiple while iterations are needed
    assert mock_prepare_batch_params.call_count == 2
    assert mock_all_fetch_batch_steps.call_count == 3
    assert generator.end_of_file is True


@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator."
       "MultithreadedBookmarkGenerator._all_fetch_batch_steps")
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.time.sleep")
def test_fetch_batch_continuously_sleep_branch(mock_time_sleep, mock_all_fetch_batch_steps):
    mock_all_fetch_batch_steps.return_value = False

    generator = MultithreadedBookmarkGeneratorFake()
    generator.buffer = [{'encoded_key': f'0-{record_no}-test'} for record_no in range(10)]
    generator.batch_limit = 5

    fetch_batch_thread = threading.Thread(target=generator.fetch_batch_continuously)
    fetch_batch_thread.start()

    while len(generator.buffer) >= generator.batch_limit:
        time.sleep(0.1)
        generator.buffer.pop()
    fetch_batch_thread.join()

    assert mock_time_sleep.call_count > generator.batch_limit


@patch.object(MultithreadedBookmarkGenerator, 'prepare_batch',
              side_effect=MultithreadedBookmarkGenerator.prepare_batch,
              autospec=True)
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.MultithreadedRequestsPool.queue_request")
def test_queue_batches(mock_queue_request,
                       mock_prepare_batch):
    mock_client = ClientMock()
    mock_endpoint_path = 'test_endpoint_path'
    mock_endpoint_api_method = 'POST'
    mock_endpoint_api_version = 'v2'
    mock_endpoint_api_key_type = 'test_endpoint_api_key_type'
    mock_endpoint_body = {'filterCriteria': {"field": "creationDate",
                                             "operator": "BETWEEN",
                                             "value": '2020-01-01T12:00:00.000000Z',
                                             "secondValue": '2020-02-01T03:01:20.000000Z'},
                          'sortingCriteria': {"field": "encodedKey",
                                              "order": "ASC"}
                          },
    mock_params = {'paginationDetails': 'OFF',
                   'detailsLevel': 'FULL'}
    mock_overlap_window = 20
    mock_batch_limit = 4000
    mock_artificial_limit = mock_client.page_size

    mock_queue_request.side_effect = [Mock() for _ in range(0, mock_batch_limit, mock_artificial_limit)]

    generator = MultithreadedBookmarkGeneratorFake(client=mock_client)
    generator.overlap_window = mock_overlap_window
    generator.batch_limit = mock_batch_limit
    generator.endpoint_path = mock_endpoint_path
    generator.endpoint_api_method = mock_endpoint_api_method
    generator.endpoint_api_version = mock_endpoint_api_version
    generator.endpoint_api_key_type = mock_endpoint_api_key_type
    generator.endpoint_body = mock_endpoint_body
    generator.params = mock_params
    generator.limit = mock_client.page_size + mock_overlap_window

    assert generator.offset == 0
    features = generator.queue_batches()

    assert len(features) == mock_batch_limit // mock_artificial_limit

    mock_params['offset'] = 0
    mock_params['limit'] = mock_artificial_limit + mock_overlap_window
    calls = []
    for offset in range(0, mock_batch_limit, mock_artificial_limit):
        calls.append(call(mock_client, 'test_stream', mock_endpoint_path, mock_endpoint_api_method,
                          mock_endpoint_api_version, mock_endpoint_api_key_type, mock_endpoint_body, dict(mock_params)))
        mock_params['offset'] += mock_artificial_limit

    # test if the queue_request method is called using the correct offset
    mock_queue_request.assert_has_calls(calls)

    # test if the methods are called the correct amount of times
    assert mock_prepare_batch.call_count == mock_batch_limit / mock_client.page_size
    assert mock_queue_request.call_count == mock_batch_limit / mock_client.page_size
    assert generator.offset == mock_batch_limit - mock_client.page_size


@patch.object(MultithreadedBookmarkGenerator, 'error_check_and_fix',
              side_effect=MultithreadedBookmarkGenerator.error_check_and_fix,
              autospec=True)
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator."
       "MultithreadedBookmarkGenerator.stop_all_request_threads")
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.MultithreadedBookmarkGenerator.transform_batch")
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.transform_json")
def test_collect_batches(mock_transform_json,
                         mock_transform_batch,
                         mock_stop_all_request_threads,
                         mock_error_check_and_fix):
    mock_batch_limit = 4000
    mock_artificial_limit = 200
    mock_overlap_window = 20
    mock_client = ClientMock(page_size=200)

    generator = MultithreadedBookmarkGeneratorFake(client=mock_client)
    generator.artificial_limit = mock_artificial_limit
    generator.batch_limit = mock_batch_limit
    generator.overlap_window = mock_overlap_window

    # generate fake data as if they were extracted from the API
    mock_records = [[HashableDict({'encoded_key': f'0-{record_no}'})
                     for record_no in range(0, mock_client.page_size + mock_overlap_window)], ]
    for batch_no in range(1, (mock_batch_limit // mock_artificial_limit) - 1):
        mock_batch = mock_records[batch_no - 1][-mock_overlap_window:]
        for record_no in range(0, mock_client.page_size):
            mock_batch.append(HashableDict({'encoded_key': f'{batch_no}-{record_no}'}))
        mock_records.append(mock_batch)
    mock_records.append([])

    mock_transform_batch.side_effect = mock_records

    mock_features = [Mock() for _ in range(0, mock_batch_limit, mock_artificial_limit)]
    buffer, stop_iteration = generator.collect_batches(mock_features)

    assert stop_iteration is True
    assert all(record in buffer for batch in mock_records[:-1] for record in batch)

    assert mock_transform_json.call_count == mock_batch_limit / mock_artificial_limit
    assert mock_transform_batch.call_count == mock_batch_limit / mock_artificial_limit
    # -2 because the first call is skipped and the last future value is an empty list
    assert mock_error_check_and_fix.call_count == (mock_batch_limit / mock_artificial_limit) - 2
    mock_stop_all_request_threads.assert_called_once()


@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.transform_json")
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.time.sleep")
def test_collect_batches_thread_not_done(mock_time_sleep, mock_transform_json):
    def create_mock_future_pending():
        mock_future_pending = Mock()
        mock_future_pending.done = Mock()
        mock_future_pending.done.side_effect = [False, False, True]
        return mock_future_pending

    mock_future_done = Mock()
    mock_future_done.done = Mock()
    mock_future_done.done.return_value = True

    generator = MultithreadedBookmarkGeneratorFake()

    mock_futures = [mock_future_done for _ in
                    range(0, generator.batch_limit - generator.artificial_limit,
                          generator.artificial_limit)] + \
                   [create_mock_future_pending()]

    generator.collect_batches(mock_futures)

    # test if the sleep function is called when the request isn't finished
    assert mock_time_sleep.call_count == 2

    # just a check to make sure the algorithm passed over the while loop
    mock_transform_json.assert_called()


@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.datetime_to_tz")
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator.str_to_localized_datetime")
@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator.preprocess_record")
@patch("tap_mambu.tap_generators.multithreaded_bookmark_generator."
       "MultithreadedBookmarkGenerator.set_intermediary_bookmark")
def test_preprocess_record(mock_set_intermediary_bookmark, mock_offset_preprocess_record,
                           mock_str_to_localized_datetime, mock_datetime_to_tz):
    mock_bookmark_field = 'test_field'
    generator = MultithreadedBookmarkGeneratorFake()
    generator.endpoint_bookmark_field = mock_bookmark_field

    # test the behaviour with a record that doesn't contain the bookmarked field
    mock_record = {'encoded_key': 'test'}
    mock_offset_preprocess_record.return_value = mock_record
    mock_datetime_to_tz.return_value = mock_record
    mock_str_to_localized_datetime.return_value = mock_record

    generator.preprocess_record(mock_record)
    mock_str_to_localized_datetime.assert_not_called()
    mock_datetime_to_tz.assert_not_called()
    mock_set_intermediary_bookmark.assert_not_called()

    # test the behaviour with a record that does contain the bookmarked field
    mock_record = {'encoded_key': 'test',
                   'test_field': '2022-01-01T00:00:00Z+03:00'}
    datetime_in_local_tz = datetime.now(timezone("Europe/Bucharest")).replace(year=2022, month=1, day=1, hour=3,
                                                                              minute=0, second=0, microsecond=0)
    datetime_in_utc_tz = datetime.utcnow().replace(year=2022, month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
    mock_offset_preprocess_record.return_value = mock_record
    mock_str_to_localized_datetime.return_value = datetime_in_local_tz
    mock_datetime_to_tz.return_value = datetime_in_utc_tz

    generator.preprocess_record(mock_record)
    mock_str_to_localized_datetime.assert_called_with(mock_record[mock_bookmark_field])
    mock_datetime_to_tz.assert_called_with(datetime_in_local_tz, "UTC")
    mock_set_intermediary_bookmark.assert_called_with(datetime_in_utc_tz)


@patch("tap_mambu.tap_generators.multithreaded_offset_generator."
       "MultithreadedOffsetGenerator.preprocess_batches")
def test_preprocess_batches(mock_offset_preprocess_batches):
    mock_buffer = [{'encoded_key': f'value_{no}'} for no in range(100)]
    mock_offset_preprocess_batches.return_value = mock_buffer

    generator = MultithreadedBookmarkGeneratorFake()
    generator.preprocess_batches(mock_buffer)
    assert generator.last_batch_size == len(mock_buffer)



from .processor import TapProcessor


class ChildProcessor(TapProcessor):
    def __init__(self, catalog, stream_name, client, config, state, sub_type, generators, parent_id):
        super().__init__(catalog, stream_name, client, config, state, sub_type, generators)
        self.endpoint_parent_id = parent_id

    def _init_endpoint_config(self):
        super()._init_endpoint_config()
        self.endpoint_parent = 'parent'

    def process_record(self, record, time_extracted, bookmark_field):
        record[f'{self.endpoint_parent}_id'] = self.endpoint_parent_id
        return super().process_record(record, time_extracted, bookmark_field)



from mock import Mock, patch
from time import sleep
from copy import deepcopy

from tap_mambu.helpers.multithreaded_requests import MultithreadedRequestsPool, PerformanceMetrics


@patch("tap_mambu.helpers.multithreaded_requests.ThreadPoolExecutor.submit")
def test_queue_request(mock_thread_pool_exec_submit):
    MultithreadedRequestsPool.queue_request('client', 'stream_name', 'stream_path', 'GET', 'v2', '', {}, {})

    # test if the submit params are the same as queue_request params
    mock_thread_pool_exec_submit.assert_called_once_with(MultithreadedRequestsPool.run, 'client', 'stream_name',
                                                         'stream_path', 'GET', 'v2', '', {}, {})


@patch("tap_mambu.helpers.client.MambuClient.request")
@patch("tap_mambu.helpers.multithreaded_requests.PerformanceMetrics", wraps=PerformanceMetrics)
@patch("tap_mambu.helpers.multithreaded_requests.MultithreadedRequestsPool.run", wraps=MultithreadedRequestsPool.run)
def test_run(mock_multithreaded_request_pool_run, mock_performance_metrics, mock_client_request):
    request_result = [{'status_code': 200,
                       'content': {'test': 'result'}},
                      {'status_code': 400,
                       'content': {'test': ''}},
                      {'status_code': 500,
                       'content': {}},
                      ]
    mock_client = Mock()
    mock_client.request = mock_client_request
    mock_client.request.side_effect = deepcopy(request_result)

    client = mock_client
    stream_names = ['test_stream_name1', 'test_stream_name2', '']
    endpoint_paths = ['test_stream_path1', 'test_stream_path2', '']
    api_methods = ['GET', 'POST', '']
    endpoint_api_version = 'v2'
    endpoint_api_key_type = ''
    endpoint_body = [{'filterCriteria': {"field": "lastModifiedDate",
                                         "operator": "AFTER",
                                         "value": ''}
                      },
                     {'filterCriteria': {"field": "creationDate",
                                         "operator": "BETWEEN",
                                         "value": '2020-01-01T12:00:00.000000Z',
                                         "secondValue": '2020-02-01T03:01:20.000000Z'},
                      'sortingCriteria': {"field": "encodedKey",
                                          "order": "ASC"}
                      },
                     {'filterCriteria': {"field": "",
                                         "operator": "",
                                         "value": ''}
                      }
                     ]
    endpoint_params = [{'offset': 500, 'limit': 510},
                       {'offset': 0, 'limit': 0},
                       {'offset': '', 'limit': ''}]

    futures = []
    for idx, stream_name in enumerate(stream_names):
        futures.append(
            MultithreadedRequestsPool.queue_request(client, stream_name, endpoint_paths[idx], api_methods[idx],
                                                    endpoint_api_version, endpoint_api_key_type,
                                                    endpoint_body, endpoint_params[idx]))
    for idx, future in enumerate(futures):
        while not future.done():
            sleep(0.001)

        # test that the request response isn't modified
        assert future.result() == request_result[idx]

    for idx, stream_name in enumerate(stream_names):
        # test if the run method is called with the correct params
        mock_multithreaded_request_pool_run.assert_any_call(client, stream_name,
                                                            endpoint_paths[idx], api_methods[idx],
                                                            endpoint_api_version, endpoint_api_key_type,
                                                            endpoint_body, endpoint_params[idx])
        # test if the client request method is called with the correct params
        mock_client_request.assert_any_call(method=api_methods[idx],
                                            path=endpoint_paths[idx],
                                            version=endpoint_api_version,
                                            apikey_type=endpoint_api_key_type,
                                            params=f'offset={endpoint_params[idx]["offset"]}&limit={endpoint_params[idx]["limit"]}',
                                            endpoint=stream_name,
                                            json=endpoint_body)

    # test if the performance metrics collector is called
    mock_performance_metrics.assert_called_with(metric_name='generator')



config_json = {'username': 'unit',
               'password': 'test',
               'apikey': '',
               'subdomain': 'unit.test',
               'test_full_url': 'http://127.0.0.1:1080/test',
               'start_date': '2021-06-01T00:00:00Z',
               'lookback_window': 30,
               'user_agent': '',
               'page_size': '200',
               'apikey_audit': 'apikey_audit_test'}



import backoff
import requests
import requests.adapters
from requests.exceptions import ConnectionError
from singer import metrics, get_logger


LOGGER = get_logger()
class ClientError(Exception):
    """class representing Generic Http error."""

    message = None

    def __init__(self, message=None, response=None):
        super().__init__(message or self.message)
        self.response = response

class MambuError(ClientError):

    message = "Unable to process request"

class MambuBadRequestError(MambuError):

    message = "400: Unable to process request"

class MambuUnauthorizedError(MambuError):

    message = "401: Invalid credentials provided"

class MambuRequestFailedError(MambuError):

    message = "402: Unable to process request"

class MambuNotFoundError(MambuError):

    message = "404: Resource not found"

class MambuMethodNotAllowedError(MambuError):

    message = "405: Method Not Allowed"

class MambuConflictError(MambuError):

    message = "409: Conflict"

class MambuForbiddenError(MambuError):

    message = "403: Insufficient permission to access resource"

class MambuUnprocessableEntityError(MambuError):

    message = "422: Unable to process request"

class MambuApiLimitError(ClientError):

    message = "429: The API limit exceeded"

class MambuInternalServiceError(MambuError):

    message = "Server Fault, Unable to process request"

class MambuNoCredInConfig(MambuError):

    message = "Creds Not Provided"

class MambuNoSubdomainInConfig(MambuError):

    message = "Subdomain not Configured"

class MambuNoAuditApikeyInConfig(MambuError):

    message = "API Key not provided"

ERROR_CODE_EXCEPTION_MAPPING = {
    400: MambuBadRequestError,
    401: MambuUnauthorizedError,
    402: MambuRequestFailedError,
    403: MambuForbiddenError,
    404: MambuNotFoundError,
    405: MambuMethodNotAllowedError,
    409: MambuConflictError,
    422: MambuUnprocessableEntityError,
    429: MambuApiLimitError,
    500: MambuInternalServiceError}

def raise_for_error(response):
    """
    Raises the associated response exception.
    :param resp: requests.Response object
    """
    try:
        response.raise_for_status()
    except (requests.HTTPError, requests.ConnectionError) as error:
        try:
            error_code = response.status_code
            if response.status_code >= 500:
                exc = MambuInternalServiceError
                message = f"{response.status_code}: {MambuInternalServiceError.message}"
            else:
                exc, message = ERROR_CODE_EXCEPTION_MAPPING.get(error_code, MambuError), None
            try:
                if len(response.content) != 0:
                    response = response.json()
                    if ('error' in response) or ('errorCode' in response):
                        message = '%s: %s' % (response.get('error', str(error)),
                                            response.get('message', 'Unknown Error'))
            except (ValueError, TypeError):
                pass
            raise exc(message, response) from None
        except (ValueError, TypeError):
            raise MambuError(error) from error

class MambuClient(object):
    def __init__(self,
                 username,
                 password,
                 apikey,
                 subdomain,
                 apikey_audit,
                 page_size,
                 user_agent=''):
        self.__username = username
        self.__password = password
        self.__subdomain = subdomain
        base_url = "https://{}.mambu.com/api".format(subdomain)
        self.base_url = base_url
        self.page_size = page_size
        self.__user_agent = f'MambuTap-{user_agent}' if user_agent else 'MambuTap'
        self.__apikey = apikey
        self.__session = requests.Session()
        self.__adapter = requests.adapters.HTTPAdapter(pool_maxsize=100)
        self.__session.mount("https://", self.__adapter)
        self.__verified = False
        self.__apikey_audit = apikey_audit

    def __enter__(self):
        self.__verified = self.check_access()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.__session.close()

    @backoff.on_exception(backoff.expo,
                          MambuInternalServiceError,
                          max_tries=5,
                          factor=2)
    def check_access(self):
        use_apikey = False
        if self.__username is None or self.__password is None:
            if self.__apikey is None:
                raise MambuNoCredInConfig(
                    'Error: Missing username or password, or apikey in config.json.'
                )
            else:
                use_apikey = True
        if self.__subdomain is None:
            raise MambuNoSubdomainInConfig('Error: Missing subdomain in config.json.')
        headers = {}
        # Endpoint: simple API call to return a single record (org settings) to test access
        # https://support.mambu.com/docs/organisational-settings-api#get-organisational-settings
        endpoint = 'settings/organization'
        url = '{}/{}'.format(self.base_url, endpoint)
        headers['User-Agent'] = self.__user_agent
        headers['Accept'] = 'application/vnd.mambu.v1+json'
        if use_apikey:
            # Api Key API Consumer Authentication: https://support.mambu.com/docs/api-consumers
            self.__session.headers['apikey'] = self.__apikey
        else:
            # Basic Authentication: https://api.mambu.com/?http#authentication
            self.__session.auth = (self.__username, self.__password)
        response = self.__session.get(
            url=url,
            headers=headers)
        if response.status_code != 200:
            LOGGER.error('Error status_code = {}'.format(response.status_code))
            raise_for_error(response)
        else:
            return True

    @backoff.on_exception(backoff.expo,
                          (MambuInternalServiceError, ConnectionError, MambuApiLimitError),
                          max_tries=7,
                          factor=3)
    def request(self, method, path=None, url=None, json=None, version=None, apikey_type=None, **kwargs):
        if not self.__verified:
            self.__verified = self.check_access()

        if not version:
            version = 'v2'

        if not url and path:
            url = '{}/{}'.format(self.base_url, path)

        if 'endpoint' in kwargs:
            endpoint = kwargs['endpoint']
            del kwargs['endpoint']
        else:
            endpoint = None

        if 'headers' not in kwargs:
            kwargs['headers'] = {}

        # Version represents API version (e.g. v2): https://api.mambu.com/?http#versioning
        if version == 'v2':
            kwargs['headers']['Accept'] = 'application/vnd.mambu.{}+json'.format(version)

        if apikey_type == 'audit':
            if self.__apikey_audit is None:
                raise MambuNoAuditApikeyInConfig(
                    'Error: Missing apikey_audit in config.json.'
                )
            else:
                kwargs['headers']['apikey'] = self.__apikey_audit

        kwargs['headers']['User-Agent'] = self.__user_agent

        if method == 'POST':
            kwargs['headers']['Content-Type'] = 'application/json'

        with metrics.http_request_timer(endpoint) as timer:
            response = self.__session.request(
                method=method,
                url=url,
                json=json,
                **kwargs)
            timer.tags[metrics.Tag.http_status_code] = response.status_code


        if response.status_code != 200:
            raise_for_error(response)

        return response.json()



import inspect
import logging
import os
from copy import deepcopy
from datetime import datetime

from unittest.mock import MagicMock

from pytz import timezone

from tap_mambu.helpers import transform_json, convert
from tap_mambu.helpers.generator_processor_pairs import get_available_streams
from tap_mambu.tap_generators.generator import TapGenerator
from tap_mambu.tap_generators.multithreaded_bookmark_generator import MultithreadedBookmarkDayByDayGenerator, \
    MultithreadedBookmarkGenerator
from . import setup_generator_base_test
from ..constants import config_json

FIXTURES_PATH = f"{os.path.dirname(os.path.abspath(inspect.stack()[0][1]))}/fixtures"


def test_generator_fetch_batch_flow():
    client_mock = MagicMock()
    client_mock.page_size = 5
    client_mock.request = MagicMock()
    client_mock.request.side_effect = [[{"id": nr + page*5} for nr in range(5)] for page in range(3)]
    generator = TapGenerator(stream_name="loan_accounts",
                             client=client_mock,
                             config=config_json,
                             state={"currently_syncing": "loan_accounts"},
                             sub_type="self")
    generator_endpoint_api_method = generator.endpoint_api_method
    generator_endpoint_path = generator.endpoint_path
    generator_endpoint_api_version = generator.endpoint_api_version
    generator_endpoint_api_key_type = generator.endpoint_api_key_type
    generator_endpoint_body = {"sortingCriteria": generator.endpoint_sorting_criteria,
                               "filterCriteria": generator.endpoint_filter_criteria}

    for batch_index in range(3):
        generator.prepare_batch()
        generator_endpoint_querystring = '&'.join([f'{key}={value}' for (key, value) in generator.params.items()])

        assert {"detailsLevel": "FULL", "limit": 5,
                "offset": 0 + batch_index * 5, "paginationDetails": "OFF"} == generator.params
        raw_batch = generator.fetch_batch()
        client_mock.request.assert_called_with(method=generator_endpoint_api_method,
                                               path=generator_endpoint_path,
                                               version=generator_endpoint_api_version,
                                               apikey_type=generator_endpoint_api_key_type,
                                               params=generator_endpoint_querystring,
                                               endpoint="loan_accounts",
                                               json=generator_endpoint_body)

        assert raw_batch == [{"id": nr + batch_index * 5} for nr in range(5)]

        generator.buffer = transform_json(raw_batch, "loan_accounts")
        assert generator.buffer == [{"custom_fields": [], "id": nr + batch_index * 5} for nr in range(5)]

        generator.last_batch_size = len(generator.buffer)

        # Simulate end of batch, so we can fetch a new one
        generator.buffer = []
        generator.offset += generator.limit


def test_generator_iterator_flow():
    client_mock = MagicMock()
    client_mock.page_size = 5
    client_data = [[{"id": nr + page*5} for nr in range(5)] for page in range(3)]
    generator = TapGenerator(stream_name="loan_accounts",
                             client=client_mock,
                             config=config_json,
                             state={"currently_syncing": "loan_accounts"},
                             sub_type="self")
    generator.prepare_batch = MagicMock()
    generator.fetch_batch = MagicMock(side_effect=client_data)

    expected_records = [{"custom_fields": [], "id": record["id"]} for record in deepcopy(sum(client_data, []))]
    record_count = 0
    for record in generator:
        assert record == expected_records[record_count]
        record_count += 1


def test_transform_batch_custom_field():
    from tap_mambu.helpers import transform_json

    assert transform_json([
        {"encodedKey": "123", "creation_date": "asd",
         "_random_field": {"cevaId": "asd"},
         "_random_field_2": [{"a": "b", "c": "d"}],
         "_random_field_3": 0}
    ], "loan_accounts") == [
        {"encoded_key": "123", "creation_date": "asd", "custom_fields": [
            {"field_set_id": "_random_field", "id": "cevaId", "value": "asd"},
            {"field_set_id": "_random_field_2", "id": "a", "value": "b"},
            {"field_set_id": "_random_field_2", "id": "c", "value": "d"}
        ]}
    ]


def test_transform_batch_custom_field_auto_add():
    from tap_mambu.helpers import transform_json

    assert transform_json([
        dict(encodedKey="123", creation_date="asd", _test=2)
    ], "loan_accounts") == [
        dict(encoded_key="123", creation_date="asd", custom_fields=list())
    ]


def test_transform_batch_larger_batch():
    from tap_mambu.helpers import transform_json

    assert transform_json([
        dict(encodedKey="123", creation_date="asd", custom_fields=list()) for _ in range(10)
    ], "loan_accounts") == [
        dict(encoded_key="123", creation_date="asd", custom_fields=list()) for _ in range(10)
    ]


def test_generator_bookmark_flow():
    stream_bookmarks = {
        "activities": "timestamp",
        "branches": "lastModifiedDate",
        "centres": "lastModifiedDate",
        "clients": "lastModifiedDate",
        "communications": "creationDate",
        "credit_arrangements": "lastModifiedDate",
        "deposit_accounts": "lastModifiedDate",
        "deposit_products": "lastModifiedDate",
        "gl_accounts": "lastModifiedDate",
        "gl_journal_entries": "creationDate",
        "groups": "lastModifiedDate",
        "deposit_transactions": "creationDate",
        "installments": "lastPaidDate",
        "interest_accrual_breakdown": "creationDate",
        "loan_products": "lastModifiedDate",
        "loan_transactions": "creationDate",
        "tasks": "lastModifiedDate",
        "users": "lastModifiedDate",
    }
    stream_custom_date = {
        # Activities stream needs the 'activity' field, which contains the actual record
        # This stream is also iterated in DESC order, so the bookmark will result in the lowest date (6th, not 7th)
        "activities": [{
            "client": "N/A", "activity": {
                "id": index, "timestamp": f"2022-06-06T00:00:00.{index:06d}Z-07:00"}
        } for index in range(400)] + [{
            "client": "N/A", "activity": {
                "id": index, "timestamp": f"2022-06-07T00:00:00.{index:06d}Z-07:00"}
        } for index in range(400)],
        # "branches": ["lastModifiedDate"],
        # "centres": ["lastModifiedDate"],
        # "clients": ["lastModifiedDate"],
        # "communications": ["creationDate"],
        # "credit_arrangements": ["lastModifiedDate"],
        # "deposit_accounts": ["lastModifiedDate"],
        # "deposit_products": ["lastModifiedDate"],
        # "gl_accounts": ["lastModifiedDate"],
        # "gl_journal_entries": ["creationDate"],
        # "groups": ["lastModifiedDate"],
        # "deposit_transactions": ["creationDate"],
        # "installments": ["lastPaidDate"],
        # "interest_accrual_breakdown": ["creationDate"],
        # "loan_products": ["lastModifiedDate"],
        # "loan_transactions": ["creationDate"],
        # "tasks": ["lastModifiedDate"],
        # "users": ["lastModifiedDate"],
    }
    for stream_name in stream_bookmarks:
        generators = setup_generator_base_test(stream_name, with_data=True,
                                               bookmark_field=stream_bookmarks[stream_name],
                                               custom_data=stream_custom_date.get(stream_name, None))
        for generator in generators:
            max_bookmark = None
            for record in generator:
                compare_function = lambda a, b: a > b
                if hasattr(generator, "compare_bookmark_values"):
                    compare_function = generator.compare_bookmark_values
                record_date = record[convert(stream_bookmarks[stream_name])]
                if max_bookmark is None or compare_function(record_date, max_bookmark):
                    max_bookmark = record_date

            if isinstance(generator, MultithreadedBookmarkDayByDayGenerator):
                assert 400 == generator.endpoint_intermediary_bookmark_offset, generator
                assert datetime(year=2022, month=6, day=6,
                                hour=0, minute=0, second=0,
                                microsecond=0, tzinfo=timezone("UTC")
                                ) == generator.endpoint_intermediary_bookmark_value, generator
            elif isinstance(generator, MultithreadedBookmarkGenerator):
                assert 1 == generator.endpoint_intermediary_bookmark_offset, generator
                assert datetime(year=2022, month=6, day=6,
                                hour=7, minute=0, second=0,
                                microsecond=399, tzinfo=timezone("UTC")
                                ) == generator.endpoint_intermediary_bookmark_value, generator

            if stream_name == "activities":  # Activities stream has descending order
                assert '2022-06-06T00:00:00.000000Z-07:00' == max_bookmark
            else:
                assert '2022-06-06T00:00:00.000399Z-07:00' == max_bookmark




from .child_processor import ChildProcessor


class DepositCardsProcessor(ChildProcessor):
    def _init_endpoint_config(self):
        super()._init_endpoint_config()
        self.endpoint_id_field = "reference_token"
        self.endpoint_parent = "deposit"



#!/usr/bin/env python3

import sys
import json
import singer

from tap_mambu.helpers.constants import DEFAULT_PAGE_SIZE
from tap_mambu.helpers.client import MambuClient
from tap_mambu.helpers.discover import discover
from tap_mambu.sync import sync_all_streams

LOGGER = singer.get_logger()

REQUIRED_CONFIG_KEYS = [
    'subdomain',
    'start_date',
    'user_agent'
]


def do_discover():
    LOGGER.info('Starting discover')
    catalog = discover()
    json.dump(catalog.to_dict(), sys.stdout, indent=2)
    LOGGER.info('Finished discover')


@singer.utils.handle_top_exception(LOGGER)
def main():
    parsed_args = singer.utils.parse_args(REQUIRED_CONFIG_KEYS)

    with MambuClient(parsed_args.config.get('username'),
                     parsed_args.config.get('password'),
                     parsed_args.config.get('apikey'),
                     parsed_args.config['subdomain'],
                     parsed_args.config.get('apikey_audit'),
                     int(parsed_args.config.get('page_size', DEFAULT_PAGE_SIZE)),
                     user_agent=parsed_args.config['user_agent']) as client:

        state = {}
        if parsed_args.state:
            state = parsed_args.state

        if parsed_args.discover:
            do_discover()
        elif parsed_args.catalog:
            sync_all_streams(client=client,
                             config=parsed_args.config,
                             catalog=parsed_args.catalog,
                             state=state)


if __name__ == '__main__':
    main()



# pragma pylint: disable=protected-access
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

from .perf_metrics import PerformanceMetrics as PerfMetrics


# noinspection PyProtectedMember
def show_thread_graph():
    all_timestamps = [(*generator_time, "r", "Generator") for generator_time in PerfMetrics._metrics["generator"]] + \
                     [(*processor_time, "b", "Processor") for processor_time in PerfMetrics._metrics["processor"]]
    counter = 0
    total_time = 0
    for timestamp in sorted(all_timestamps, key=lambda ts: ts[0]):
        start_time = round(timestamp[0] - PerfMetrics._metrics_start_time, 1)
        end_time = round(timestamp[1] - PerfMetrics._metrics_start_time, 1)
        plt.plot([start_time, end_time],
                 [counter, counter], color=timestamp[2], label=timestamp[3], linestyle="-")
        counter += 1
        if end_time > total_time:
            total_time = end_time
    plt.title(f"Total execution time: {total_time}s")
    plt.ylabel("Timestamp")
    plt.legend([Line2D([0], [0], color="r", lw=4), Line2D([0], [0], color="b", lw=4)], ["Generator", "Processor"])

    plt.show()


# noinspection PyProtectedMember
def show_request_duration_graph():
    data_points = [record[1] - record[0] for record in PerfMetrics._metrics["generator"]]
    x = list(range(len(data_points)))
    plt.bar(x, data_points)
    plt.show()


from .no_pagination_generator import NoPaginationGenerator


class IndexRateSourcesGenerator(NoPaginationGenerator):
    def _init_endpoint_config(self):
        super(IndexRateSourcesGenerator, self)._init_endpoint_config()
        self.endpoint_path = "indexratesources"
        self.endpoint_api_method = "GET"
        self.endpoint_sorting_criteria = {}
        self.endpoint_filter_criteria = []
        self.endpoint_bookmark_field = ""



from .generator import TapGenerator


class BranchesGenerator(TapGenerator):
    def _init_endpoint_config(self):
        super(BranchesGenerator, self)._init_endpoint_config()
        self.endpoint_path = "branches"
        self.endpoint_api_method = "GET"
        self.endpoint_params["sortBy"] = "lastModifiedDate:ASC"
        self.endpoint_bookmark_field = "lastModifiedDate"



import time
from copy import deepcopy
from threading import Thread

import backoff
from singer import get_logger

from .generator import TapGenerator
from ..helpers import transform_json
from ..helpers.multithreaded_requests import MultithreadedRequestsPool
from ..helpers.perf_metrics import PerformanceMetrics

LOGGER = get_logger()


class MultithreadedOffsetGenerator(TapGenerator):
    def _init_params(self):
        self.time_extracted = None
        self.static_params = dict(self.endpoint_params)
        self.offset = 0
        self.overlap_window = 20
        self.artificial_limit = self.client.page_size
        self.limit = self.client.page_size + self.overlap_window
        self.batch_limit = self.max_threads * self.client.page_size + self.overlap_window
        self.params = self.static_params

    def _init_config(self):
        super(MultithreadedOffsetGenerator, self)._init_config()
        self.end_of_file = False
        self.fetch_batch_thread = None
        self.last_batch_set = set()
        self.max_threads = 20

    @staticmethod
    def check_and_get_set_reunion(a: set, b: set, lower_limit: int):
        reunion = a | b
        if len(reunion) == len(a) + len(b):
            # Raise a runtime error to be caught at the top level function _all_fetch_batch_steps,
            # in order to retry extraction or stop trying after a few retries
            raise RuntimeError("Retrying extraction for last multithreaded batches as a discrepancy has been detected "
                               "between two consecutive batches. (Records have shifted due to insertions/changes/deletions)")
        if len(a) + lower_limit < len(reunion) < len(a) + len(b):
            LOGGER.warning("Discrepancies detected for last multithreaded batches extraction, but they are correctable."
                           " (Records have shifted due to insertions/changes/deletions)")
        return reunion

    @staticmethod
    def stop_all_request_threads(futures):
        for future in futures:
            future.cancel()

        for future in futures:
            while not future.done():  # Both finished and cancelled futures return 'done'
                time.sleep(0.1)

    def fetch_batch_continuously(self):
        while not self.end_of_file:
            if not self._all_fetch_batch_steps():
                self.end_of_file = True
            time.sleep(0.1)

    def queue_batches(self):
        # prepare batches (with self.limit for each of them until we reach batch_limit)
        futures = list()
        while len(self.buffer) + len(futures) * self.limit <= self.batch_limit:
            self.prepare_batch()
            # send batches to multithreaded_requests_pool
            futures.append(MultithreadedRequestsPool.queue_request(self.client, self.stream_name,
                                                                   self.endpoint_path, self.endpoint_api_method,
                                                                   self.endpoint_api_version,
                                                                   self.endpoint_api_key_type,
                                                                   deepcopy(self.endpoint_body),
                                                                   deepcopy(self.params)))
            self.offset += self.artificial_limit
        return futures

    def collect_batches(self, futures):
        # wait for responses, and check them for errors
        last_batch = set()
        final_buffer = self.last_batch_set
        stop_iteration = False
        for future in futures:
            while not future.done():
                time.sleep(0.1)
            result = future.result()
            transformed_batch = self.transform_batch(transform_json(result, self.stream_name))
            temp_buffer = set(transformed_batch)

            if not temp_buffer:  # We finished the data to extract, time to stop
                self.stop_all_request_threads(futures)
                stop_iteration = True
                break

            last_batch = temp_buffer

            if not final_buffer:
                final_buffer = final_buffer | temp_buffer
                continue

            final_buffer = self.error_check_and_fix(final_buffer, temp_buffer, futures)

            if stop_iteration:
                break

        final_buffer -= self.last_batch_set
        self.last_batch_set = last_batch

        return final_buffer, stop_iteration

    def preprocess_record(self, raw_record):
        self.buffer.append(raw_record)
        return raw_record

    def preprocess_batches(self, final_buffer):
        # if no errors found:
        # dump data into buffer
        for raw_record in final_buffer:
            self.preprocess_record(raw_record)
        self.last_batch_size = len(self.last_batch_set)

    @backoff.on_exception(backoff.expo, RuntimeError, max_tries=5)
    def _all_fetch_batch_steps(self):
        futures = self.queue_batches()
        final_buffer, stop_iteration = self.collect_batches(futures)
        self.preprocess_batches(final_buffer)
        if not final_buffer or stop_iteration:
            return False
        return True

    def error_check_and_fix(self, final_buffer: set, temp_buffer: set, futures: list):
        try:
            final_buffer = self.check_and_get_set_reunion(final_buffer, temp_buffer, self.artificial_limit)
        except RuntimeError:  # if errors are found
            LOGGER.exception("Discrepancies found in extracted data, and errors couldn't be corrected."
                             "Shutting down all remaining threads for this batch's extraction to retry it.")

            # wait all threads to finish/cancel all threads
            self.stop_all_request_threads(futures)
            LOGGER.info("Thread shutdown complete! Retrying extraction from last bookmark...")
            # retry the whole process (using backoff decorator, so we need to propagate the exception)
            # effectively rerunning this function with the same parameters
            raise
        return final_buffer

    def __iter__(self):
        if self.fetch_batch_thread is None:
            self.fetch_batch_thread = Thread(target=self.fetch_batch_continuously, name="FetchContinuouslyThread")
            self.fetch_batch_thread.start()
        return self

    def next(self):
        if not self.buffer and not self.end_of_file:
            with PerformanceMetrics(metric_name="processor_wait"):
                while not self.buffer and not self.end_of_file:
                    time.sleep(0.01)
        if not self.buffer and self.end_of_file:
            raise StopIteration()
        return self.buffer.pop(0)

    def fetch_batch(self):
        raise DeprecationWarning("Function is being deprecated, and not implemented in this subclass!")



from .generator import TapGenerator
from ..helpers import get_bookmark
from ..helpers.datetime_utils import str_to_datetime, datetime_to_utc_str, utc_now


class AuditTrailGenerator(TapGenerator):
    def _init_endpoint_config(self):
        super(AuditTrailGenerator, self)._init_endpoint_config()
        self.endpoint_path = "v1/events"
        self.endpoint_api_method = "GET"
        self.endpoint_api_version = "v1"
        self.endpoint_api_key_type = "audit"
        self.endpoint_bookmark_field = "occurred_at"

        audit_trail_bookmark = get_bookmark(self.state, self.stream_name, self.sub_type, self.start_date)
        self.audit_trail_offset = 0

        # Backwards compatibility check, such that it doesn't fail when we use a bookmark from an older Tap version
        if type(audit_trail_bookmark) is list:
            if len(audit_trail_bookmark) != 2:
                raise ValueError("Cannot parse audit trail bookmark from list because we expect 2 values!")
            self.audit_trail_offset = audit_trail_bookmark[1]
            audit_trail_bookmark = audit_trail_bookmark[0]

        self.endpoint_params = {
            "sort_order": "asc",
            "occurred_at[gte]": datetime_to_utc_str(str_to_datetime(audit_trail_bookmark)),
            "occurred_at[lte]": datetime_to_utc_str(utc_now()),
        }

    def _init_params(self):
        super(AuditTrailGenerator, self)._init_params()
        # Initializing offset with value from bookmark, so we skip N records that were extracted after the bookmark
        self.offset = self.audit_trail_offset

    def next(self):
        if not self.buffer:
            if self.last_batch_size < self.limit:
                raise StopIteration()
            # Overriding offset, as every subsequent request will include exactly one record that was already extracted
            self.offset = 1
            self._all_fetch_batch_steps()
            if not self.buffer:
                raise StopIteration()
        return self.buffer.pop(0)

    def prepare_batch(self):
        # Overriding params, as they have different names in audit trail
        # WARNING: self.static_params will be changed by the processor dinamically!
        self.params = {
            "from": self.offset,
            "size": self.limit,
            **self.static_params
        }

    def fetch_batch(self):
        # Extracting "events" field from batch, as we are returned a dict instead of a list of records
        return super(AuditTrailGenerator, self).fetch_batch()["events"]



from .multithreaded_bookmark_generator import MultithreadedBookmarkDayByDayGenerator
from ..helpers import get_bookmark
from ..helpers.datetime_utils import datetime_to_utc_str, str_to_localized_datetime, utc_now


class ActivitiesGenerator(MultithreadedBookmarkDayByDayGenerator):
    def _init_endpoint_config(self):
        super(ActivitiesGenerator, self)._init_endpoint_config()
        self.endpoint_path = "activities"
        self.endpoint_api_method = "GET"
        self.endpoint_api_version = "v1"

        self.endpoint_params["from"] = datetime_to_utc_str(str_to_localized_datetime(
                    get_bookmark(self.state, self.stream_name, self.sub_type, self.start_date)))[:10]
        self.endpoint_params["to"] = datetime_to_utc_str(utc_now())[:10]
        self.endpoint_bookmark_field = "timestamp"

    @staticmethod
    def unpack_activity(record):
        record.update(record["activity"])
        del record["activity"]
        return record

    def transform_batch(self, batch):
        return list(map(self.unpack_activity, super(ActivitiesGenerator, self).transform_batch(batch)))

    def prepare_batch_params(self):
        super(ActivitiesGenerator, self).prepare_batch_params()
        self.static_params['to'] = datetime_to_utc_str(self.endpoint_intermediary_bookmark_value)[:10]

    def compare_bookmark_values(self, a, b):
        return a < b



from . import setup_generator_base_test


def test_credit_arrangements_generator_endpoint_config_init():
    generators = setup_generator_base_test("credit_arrangements")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == 'creditarrangements'
    assert generator.endpoint_api_method == "GET"
    assert generator.endpoint_params == {
        "sortBy": "creationDate:ASC",
        "detailsLevel": "FULL",
        "paginationDetails": "OFF"
    }
    assert generator.endpoint_bookmark_field == "lastModifiedDate"



from . import setup_processor_base_test


def test_deposit_accounts_processor():
    processor = setup_processor_base_test("deposit_accounts")

    assert processor.endpoint_child_streams == ["cards"]
